# -*- coding: utf-8 -*-

"""
Intro
=====
Provide interface with the INRIA XML database released
by Prof. Gerard Huet
https://gitlab.inria.fr/huet/Heritage_Resources
(Thanks Prof. Huet for making the data available!)

Inspired by https://github.com/drdhaval2785/inriaxmlwrapper
(Thanks @drdhaval2785!)

@author: Avinash Varna (@avinashvarna)

Usage
=====
The InriaXMLWrapper utility class can also be used to lookup tags:

.. code:: python

    >>> from sanskrit_parser.util.inriaxmlwrapper import InriaXMLWrapper
    >>> db = InriaXMLWrapper()
    >>> db_tags = db.get_tags('hares')
    >>> tags == db_tags
    True


Command line usage
==================

::

    $ python -m sanskrit_parser.util.inriaxmlwrapper hares
    INFO:root:Pickle file found, loading at 2017-07-31 14:35:56.093000
    INFO:root:Loading finished at 2017-07-31 14:35:59.159000, took 3.066000 s
    INFO:root:Cached 666994 forms for fast lookup
    Getting tags for hares
    ('hf#1', set(['cj', 'snd', 'prim', 'para', 'md', 'sys', 'prs', 'v', 'np', 'sg', 'op']))
    ('hari#1', set(['na', 'mas', 'sg', 'gen']))
    ('hari#1', set(['na', 'mas', 'abl', 'sg']))
    ('hari#1', set(['na', 'fem', 'sg', 'gen']))
    ('hari#1', set(['na', 'fem', 'abl', 'sg']))
    ('hari#2', set(['na', 'mas', 'sg', 'gen']))
    ('hari#2', set(['na', 'mas', 'abl', 'sg']))
    ('hari#2', set(['na', 'fem', 'sg', 'gen']))
    ('hari#2', set(['na', 'fem', 'abl', 'sg']))

"""

from __future__ import print_function
import os
import logging
import importlib.resources
from collections import namedtuple

from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SCHEMES
from sanskrit_parser.util.lexical_lookup import LexicalLookup
from sanskrit_parser.util.inriatagmapper import inriaTagMapper

try:
    import cPickle as pickle
except ImportError:
    import pickle

import sqlite3


_db = namedtuple('_db', ['db_file', 'tags', 'stems', 'buf'])


class InriaXMLWrapper(LexicalLookup):
    """
    Class to interface with the INRIA XML database released
    by Prof. Gerard Huet
    https://gitlab.inria.fr/huet/Heritage_Resources
    """

    '''
    The custom database format has two parts:
        1. A pickle file that contains a list of stems,
           a list of tags, and a serialized buffer of the
           indices of stems and tags for each form. The indices
           are used as it is more efficient to store such integers
           instead of the string for each tag.
        2. An sqlite file that maps each form to the position
           within the buffer that contains the serialized tuple
           of stems and tags for that form. An sqlite database
           is used to avoid having to build a huge dict in
           memory for the 600K forms that are present in this db,
           which consumes a lot of memory. (See
           https://github.com/kmadathil/sanskrit_parser/issues/151)

    To lookup the tag for a form, we use the sqlite db to find the
    position in the buffer, deserialize the data at that position,
    which gives us a list of the tag set for that form. For each
    item in that list, we then lookup the right stem and tag in
    the list of stems and tags loaded from the pickle file
    '''

    def __init__(self, logger=None):
        self.pickle_file = "inria_forms.pickle"
        self.logger = logger or logging.getLogger(__name__)
        with importlib.resources.path('sanskrit_parser', 'data') as base_dir:
            db_file = os.path.join(base_dir, "inria_forms_pos.db")
            pkl_path = os.path.join(base_dir, "inria_stems_tags_buf.pkl")
        self.db = self._load_db(db_file, pkl_path)

    @staticmethod
    def _load_db(db_file, pkl_path):
        with open(pkl_path, 'rb') as f:
            stems = pickle.load(f)
            tags = pickle.load(f)
            buf = f.read()
        db = _db(db_file, tags, stems, buf)
        return db

    def _get_tags(self, word):
        db = self.db
        conn = sqlite3.connect(db.db_file)
        cursor = conn.cursor()
        res = cursor.execute('SELECT * FROM forms WHERE form=?', (word,)).fetchone()
        if res is None:
            return None
        pos = res[1]
        tag_index_list = pickle.loads(db.buf[pos:])
        tags = []
        for tag_index in tag_index_list:
            tags.append(self._decode_tags(tag_index, db.tags, db.stems))
        return tags

    @staticmethod
    def _decode_tags(tag_index, tags, stems):
        t = [tags[x] for x in tag_index[1]]
        stem = stems[tag_index[0]]
        return (stem, set(t))

    def valid(self, word):
        conn = sqlite3.connect(self.db.db_file)
        cursor = conn.cursor()
        res = cursor.execute('SELECT COUNT(1) FROM forms WHERE form = ?', (word,)).fetchone()
        return res[0] > 0

    def get_tags(self, word, tmap=True):
        tags = self._get_tags(word)
        if tmap and (tags is not None):
            tags = inriaTagMapper(tags)
        return tags


if __name__ == "__main__":
    from argparse import ArgumentParser

    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = ArgumentParser(description='Interface to INRIA XML database')
        # Input Encoding (autodetect by default)
        parser.add_argument('--input-encoding', type=str, default=None)
        parser.add_argument('--loglevel', type=str, default="info",
                            help="logging level. Can be any level supported by logging module")

        parser.add_argument('word', nargs='?', type=str,
                            default=None,
                            help="Word to look up")
        parser.add_argument('--no-map-tags', dest='map_tags',
                            action='store_false')
        return parser.parse_args()

    def main():
        args = getArgs()
        if args.input_encoding is None:
            ie = None
        else:
            ie = SCHEMES[args.input_encoding]

        if args.loglevel:
            numeric_level = getattr(logging, args.loglevel.upper(), None)
            if not isinstance(numeric_level, int):
                raise ValueError('Invalid log level: %s' % args.loglevel)
            logging.basicConfig(level=numeric_level)

        word_in = SanskritImmutableString(args.word, encoding=ie).canonical()
        xmlDB = InriaXMLWrapper()
        print("Getting tags for", word_in)
        tags = xmlDB.get_tags(word_in, tmap=args.map_tags)
        if tags is not None:
            for t in tags:
                print(t)

    main()



# -*- coding: utf-8 -*-
"""
@author: Karthik Madathil (github: @kmadathil)

"""
from argparse import ArgumentParser, Action
import logging
from sanskrit_parser.base.sanskrit_base import SLP1, DEVANAGARI
from sanskrit_parser.generator.paninian_object import PaninianObject
from sanskrit_parser.generator.prakriya import Prakriya, PrakriyaVakya
from sanskrit_parser.generator.pratyaya import *  # noqa: F403
from sanskrit_parser.generator.dhatu import *  # noqa: F403
from sanskrit_parser.generator.pratipadika import *  # noqa: F403
from sanskrit_parser.generator.sutras_yaml import sutra_list
from sanskrit_parser import enable_file_logger, enable_console_logger

logger = logging.getLogger(__name__)


def run_pp(s, verbose=False):
    pl = []
    # Assemble list of inputs
    for i in range(len(s)):
        def _gen_obj(s, i):
            if isinstance(s[i], tuple) or isinstance(s[i], list):
                lelem = [_gen_obj(s[i], ii) for (ii, ss) in enumerate(s[i])]
            else:
                lelem = s[i]
            return lelem
        lelem = _gen_obj(s, i)
        pl.append(lelem)
    p = Prakriya(sutra_list, PrakriyaVakya(pl))
    p.execute()
    if verbose:
        p.describe()
    o = p.output()
    return o


# Insert all sup vibhaktis one after the other, with avasAnas
# Return results with avasAnas stripped as 8x3 list of lists
def generate_vibhakti(pratipadika, verbose=False):
    r = []
    for ix, s in enumerate(sups):  # noqa: F405
        if verbose:
            logger.info(f"Vibhakti {ix+1} {s}")
        else:
            logger.debug(f"Vibhakti {ix+1} {s}")
        r.append([])
        for jx, ss in enumerate(s):
            # For nitya eka/dvi/bahuvacana, generate only the appropriate
            if (((jx == 0) and pratipadika.hasTag("nityEkavacana")) or
                ((jx == 1) and pratipadika.hasTag("nityadvivacana")) or
                ((jx == 2) and pratipadika.hasTag("nityabahuvacana")) or
                (not (pratipadika.hasTag("nityEkavacana") or
                      pratipadika.hasTag("nityadvivacana") or
                      pratipadika.hasTag("nityabahuvacana")))):
                t = [(pratipadika, ss), avasAna]  # noqa: F405
                _r = run_pp(t, verbose)
                r[-1].append(_r)
                p = [''.join([str(x) for x in y]) for y in _r]
                pp = ", ".join([x.strip('.') for x in p])
                if verbose:
                    logger.info(f"Vacana {jx+1} {ss} {pp}")
                else:
                    logger.debug(f"Vacana {jx+1} {ss} {pp}")
    return r


last_option = False


class CustomAction(Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        # if nargs is not None:
        #   raise ValueError("nargs not allowed")
        super(CustomAction, self).__init__(option_strings, dest, nargs, **kwargs)
        logger.debug(f"Initializing CustomAction {option_strings}, {dest}")

    def __call__(self, parser, namespace, values, option_string=None):
        logger.debug('%r %r %r' % (namespace, values, option_string))
        global last_option
        assert not last_option, f"Option {option_string} added after avasana"
        if getattr(namespace, self.dest) is None:
            _n = []
            # This tracks the hierarchical input list
            setattr(namespace, self.dest, _n)
            # Last item of this is always the current level of the input
            setattr(namespace, "pointer", [_n])
        if values is not None:
            if isinstance(values, str):
                values = [values]
            for v in values:
                assert v in globals(), f"{v} is not defined!"
                getattr(namespace, "pointer")[-1].append(globals()[v])
        else:
            if option_string == "-o":  # Open
                _l = []
                # Add a new level at the end of current list
                getattr(namespace, "pointer")[-1].append(_l)
                # Designate new list as current list
                getattr(namespace, "pointer").append(_l)
            elif option_string == "-c":  # Close
                # Current is updated to previous
                getattr(namespace, "pointer").pop()
            elif option_string == "-a":  # AvasAna
                # Add avasana
                lav = getattr(namespace, self.dest)
                setattr(namespace, self.dest, [lav, avasAna])  # noqa: F405
                last_option = True
            else:
                logger.error(f"Unrecognized Option {option_string}")


class CustomActionString(Action):
    def __init__(self, option_strings, dest, nargs=None, encoding=SLP1, **kwargs):
        # if nargs is not None:
        #   raise ValueError("nargs not allowed")
        self.encoding = encoding
        super(CustomActionString, self).__init__(option_strings, dest, nargs, **kwargs)
        logger.debug(f"Initializing CustomAction {option_strings}, {dest}")

    def __call__(self, parser, namespace, values, option_string=None):
        global last_option
        assert not last_option, f"Option {option_string} added after avasana"
        encoding = self.encoding

        def _exec(value):
            # Shortcuts for two input tests not using predefined objects
            # If a string in the first place ends with * it's an anga
            # Else it's a pada
            # For everything else, use predefined objects
            if (value[-1] == "*"):
                value = value[:-1]
                value = PaninianObject(value, encoding)  # noqa: F405
                value.setTag("aNga")
            elif (value[-1] == "_"):
                value = value[:-1]
                value = PaninianObject(value, encoding)  # noqa: F405
                value.setTag("pada")
            else:
                value = PaninianObject(value, encoding)  # noqa: F405
            getattr(namespace, "pointer")[-1].append(value)

            logger.info('%r %r %r' % (namespace, values, option_string))
        if getattr(namespace, self.dest) is None:
            _n = []
            # This tracks the hierarchical input list
            setattr(namespace, self.dest, _n)
            # Last item of this is always the current level of the input
            setattr(namespace, "pointer", [_n])
        if isinstance(values, list):
            for v in values:
                _exec(v)
        else:
            _exec(values)


def get_args(argv=None):
    """
      Argparse routine.
      Returns args variable
    """

    parser = ArgumentParser(description='Paninian Generator: Prakriti + Pratyaya')
    # String to encode
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('-p', '--pratyaya', nargs="+", dest="inputs", action=CustomAction)
    parser.add_argument('-d', '--dhatu', dest="inputs", action=CustomAction)
    parser.add_argument('-t', '--pratipadika', dest="inputs", action=CustomAction)
    parser.add_argument('-s', '--string', nargs="+", dest="inputs", encoding=SLP1, action=CustomActionString)
    parser.add_argument('-o', nargs="?", dest="inputs", action=CustomAction, help="Open bracket")  # Open Brace
    parser.add_argument('-c', nargs="?", dest="inputs", action=CustomAction, help="Close bracket")
    parser.add_argument('-a', nargs="?", dest="inputs", action=CustomAction, help="Avasana")
    parser.add_argument("--vibhakti", action="store_true", help="generate all vibhaktis")
    parser.add_argument("--gen-test", action="store_true", help="generate vibhakti test")
    parser.add_argument("--verbose", action="store_true", help="verbose")

    return parser.parse_args(argv)


def cmd_line():
    # Logging
    enable_console_logger()
    args = get_args()
    if args.debug:
        enable_file_logger(level=logging.DEBUG)
    logger.info(f"Inputs {args.inputs}")
    for i in args.inputs:
        def _i(x):
            if isinstance(x, list):
                for _x in x:
                    _i(_x)
            else:
                logger.info(f"{x} {x.tags}")
        _i(i)
    logger.info("End Inputs")
    if args.vibhakti:
        if ((len(args.inputs) != 1) or (not isinstance(args.inputs[0], Pratipadika))):  # noqa: F405
            logger.info(f"Need a single pratipadika for vibhaktis, got {len(args.inputs)} inputs, first one of type {type(args.inputs[0])}")
            logger.info("Simplifying")
            r = run_pp(args.inputs, args.verbose)
            logger.debug(f"Output: {[''.join([str(x) for x in y]) for y in r]}")
            assert len(r) == 1, "Got multiple outputs"
            pp = PaninianObject.join_objects(r)
            logger.info(f"Output {pp} {pp.tags}")
        else:
            pp = args.inputs[0]
        r = generate_vibhakti(pp, args.verbose)
        print("Output")
        if args.gen_test:
            rr = [[[y[0].transcoded(DEVANAGARI) for y in va] if len(va) > 1 else va[0][0].transcoded(DEVANAGARI) for va in vi] for vi in r]
            print(f"prAtipadika[\"{str(pp)}\"] = {str(pp)}")
            print(f"viBakti[\"{str(pp)}\"] = [")
            for vi in rr:
                print(f"{vi},")
            print("]")
        else:
            for ix, vi in enumerate(r):
                print(f"{', '.join(['/'.join([''.join([x.transcoded(DEVANAGARI) for x in y]).strip('।') for y in va]) for va in vi])}")
    else:
        r = run_pp(args.inputs, args.verbose)
        print(f"Output: {[''.join([str(x) for x in y]) for y in r]}")



"""
Prakriya Engine for Panini Sutras

Takes in a list of Sutras, and executes them on a bunch of inputs

Inputs are provided at prakRti + pratyayas level. A list of such combinations
is provided. (ie: this is post karaka assignment and pratyaya selection).
For each future pada (prakRti + pratyayas), rules are executed, and output
padas are generated. Then, further rules are run on the padas themselves to
produce a vakya.

@author: kmadathil
"""

from sanskrit_parser.generator.sutra import GlobalDomains
from sanskrit_parser.generator.paninian_object import PaninianObject
from copy import deepcopy, copy
import logging
logger = logging.getLogger(__name__)


class PrakriyaVakya(object):
    """
    Prakriya Vakya class

    Start with associated prakriti + pratyayas
    Assemble into padas
    Handle Pratyaya/String Agama / Lopa

    Inputs:
        v = list.
        Elements of v can be PaninianObjects or
            lists thereof
    Internal storage:
            - List of lists of PaninianObject objects
    """
    def __init__(self, v):
        # Deepcopy is required because we add tags to objects
        # during prakriya, and we do not want predefined objects getting
        # tags.
        self.v = deepcopy(list(v))

    def need_hierarchy_at(self, ix):
        return not _isScalar(self.v[ix])

    def copy_replace_at(self, ix, r):
        vc = PrakriyaVakya(self.v)
        # As above, deepcopy to prevent predefined objects getting tags
        vc.v[ix] = deepcopy(r)
        return vc

    def copy_insert_at(self, ix, r):
        vc = PrakriyaVakya(self.v)
        # As above, deepcopy to prevent predefined objects getting tags
        vc.v.insert(ix, deepcopy(r))
        return vc

    def replace_at(self, ix, r):
        # As above, deepcopy to prevent predefined objects getting tags
        self.v[ix] = deepcopy(r)
        return self

    def insert_at(self, ix, r):
        # As above, deepcopy to prevent predefined objects getting tags
        self.v.insert(ix, deepcopy(r))
        return self

    def __getitem__(self, ix):
        return self.v[ix]

    def __len__(self):
        return len(self.v)

    def __print__(self):
        return [str(x) for x in self.v]

    def __repr__(self):
        return str([str(x) for x in self.v])


class Prakriya(object):
    """
    Prakriya Class

    Inputs:
       sutra_list: list of Sutra objects
       inputs    : PrakriyaVakya object
    """
    def __init__(self, sutra_list, inputs):
        self.sutra_list = sutra_list
        self.pre_inputs = deepcopy(inputs)
        self.inputs = copy(inputs)
        self.hier_prakriyas = []
        self.need_hier = False
        # List of alternatives
        # Used only in hierarchy is needed
        self.hier_inputs = [self.inputs]
        # Assmeble hierarchical prakriya outputs into a single list
        self.hier_outputs = [[] for x in self.inputs]
        # Scan inputs for hierarchical prakriya needs
        for ix in range(len(self.inputs)):
            if self.inputs.need_hierarchy_at(ix):
                self.need_hier = True
                # hierarchy needed here
                hp = Prakriya(sutra_list,
                              PrakriyaVakya(self.inputs[ix]))
                self.hier_prakriyas.append(hp)
                # This will execute hierarchically as needed
                hp.execute()
                hpo = hp.output()
                self.hier_outputs[ix] = hpo  # accumulate hierarchical outputs
        self.tree = PrakriyaTree()
        if self.need_hier:
            for ix, ol in enumerate(self.hier_outputs):
                if ol != []:  # Hierarchy exists here
                    tmpl = []
                    # For each alternate output
                    for o in ol:
                        hobj = PaninianObject.join_objects([o])
                        # Assemble exploded list with each input
                        # replaced by multiple alternates at this
                        # index
                        for i in self.hier_inputs:
                            tmpl.append(i.copy_replace_at(ix, hobj))
                    # Replace input list with exploded list
                    # Explosion at position ix is now dealt with
                    self.hier_inputs = tmpl
                    logger.debug(f"Hier inputs after expl {ix} {self.hier_inputs}")
            # At the end of the loop above self.hier_inputs has been fully exploded
            for i in self.hier_inputs:
                _n = PrakriyaNode(i, i, "Prakriya Hierarchical Start")
                self.tree.add_node(_n, root=True)
        else:
            _n = PrakriyaNode(self.inputs, self.inputs, "Prakriya Start")
            self.tree.add_node(_n, root=True)

        self.outputs = []
        self.domains = GlobalDomains()
        self.disabled_sutras = []
        # Sliding window counter
        self.windowIdx = 0

    # pUrvaparanityAntaraNgApavAdAnamuttarottaraM balIyaH
    def sutra_priority(self, sutras: list):
        def _winner(s1, s2):
            logger.debug(f"{s1} overrides {s1.overrides}")
            logger.debug(f"{s2} overrides {s2.overrides}")
            # Apavada
            if (s2.overrides is not None) and (s1.aps in s2.overrides):
                logger.debug(f"{s2} overrides {s1}")
                return s2
            elif (s1.overrides is not None) and (s2.aps in s1.overrides):
                logger.debug(f"{s1} overrides {s2}")
                return s1
            # Nitya
            # Antaranga
            elif (s1.bahiranga < s2.bahiranga):
                logger.debug(f"{s1} antaranga {s2}")
                return s1
            elif (s2.bahiranga < s1.bahiranga):
                logger.debug(f"{s2} antaranga {s1}")
                return s2
            # samjYA before 1.4.2 vipratizeDe param kAryam
            elif (s1._aps_num < 14000) or (s2._aps_num < 14000):
                logger.debug(f"SaMjYA, lower of {s1} {s2}")
                if s1._aps_num < s2._aps_num:
                    return s1
                else:
                    return s2
            # Also handles if one sutra is spsp and one tp
            elif (s1._aps_num > 82000) or (s2._aps_num > 82000):
                logger.debug(f"Tripadi, lower of {s1} {s2}")
                if s1._aps_num < s2._aps_num:
                    return s1
                else:
                    return s2
            else:
                # Para > purva
                logger.debug(f"Sapadasaptapadi, higher of {s1} {s2}")
                if s1._aps_num > s2._aps_num:
                    return s1
                else:
                    return s2
        _s = sutras
        w = _s[0]
        for s in _s[1:]:
            w = _winner(w, s)
        return w

    def view(self, s, node, ix=0):
        """
        Current view as seen by sutra s

        """
        # Wrapper for special "siddha" situations
        def _special_siddha(a1, a2):
            # zqutva is siddha for q lopa
            if (a1 == 84041) and (a2 == 83013):
                return True
            # q, r lopa siddha for purvadirgha
            elif ((a1 == 83013) or (a1 == 83014)) and (a2 == 63111):
                return True
            else:
                return False

        if s is not None:
            aps_num = s._aps_num
        else:
            aps_num = 0
        # Default view
        _l = self.inputs
        if node is None:
            # logger.debug(f"View {l} {node}")
            return _l
        if aps_num < 82000:
            # FIXME: Only Sapadasaptapadi implemented.
            # Need to implement asiddhavat, zutvatokorasiddhaH
            # Can see the entire sapadasaptapadi
            _n = node
            while (self.tree.parent[_n] is not None) and \
                  ((_n.sutra._aps_num > 82000)
                   and not _special_siddha(_n.sutra._aps_num, aps_num)):
                _n = self.tree.parent[_n]
            _l = _n.outputs
        else:
            # Asiddha
            # Can see all outputs of sutras less than oneself
            _n = node
            while (self.tree.parent[_n] is not None) and \
                  ((_n.sutra._aps_num > aps_num)
                   and not _special_siddha(_n.sutra._aps_num, aps_num)):
                _n = self.tree.parent[_n]
            _l = _n.outputs
        if ix > (len(_l)-2):
            # Someone has inserted something this sutra can't see
            logger.debug(f"Unseen insertion? {s} {_l} {ix}")
            ix = len(_l) - 2
        return _l[ix:ix+2]

    def _exec_single(self, node):
        l = self.sutra_list  # noqa: E741
        # Sliding window, check from left
        for ix in range(len(node.outputs)-1):
            logger.debug(f"Disabled Sutras at window {ix} {[s for s in node.outputs[ix].disabled_sutras]}")
            triggered = []
            triggered = [s for s in l if ((s.aps not in node.outputs[ix].disabled_sutras)
                                          and s.isTriggered(*self.view(s, node, ix), self.domains))]
            # Break at first index from left where trigger occurs
            if triggered:
                _ix = ix
                break
        logger.debug(f"I: {node.id} {node.outputs} {[_r.tags for _r in node.outputs]} ")
        if triggered:
            ix = _ix
            logger.debug(f"Triggered rules at window {ix}")
            for t in triggered:
                logger.debug(t)
            s = self.sutra_priority(triggered)
            v = self.view(s, node, ix)
            logger.debug(f"Sutra {s} View {v} Disabled: {[s for s in v[0].disabled_sutras]}")
            assert s.aps not in v[0].disabled_sutras
            # Transformation
            r = s.operate(*v)
            r0 = r[0]
            v0 = v[0]
            # State update
            r = s.update(*v, *r, self.domains)
            r = s.insert(*r)
            logger.debug(f"I (post update): {node.id} {node.outputs} {[_r.tags for _r in node.outputs]} ")
            logger.debug(f"I (post update): {v}")
            # Using sutra id in the disabled list to get round paninian object deepcopy
            r0.disabled_sutras.append(s.aps)
            if s.optional:
                # Prevent optional sutra from executing on the same node again
                v0.disabled_sutras.append(s.aps)
            # Overridden sutras disabled
            if s.overrides is not None:
                for so in l:
                    if so.aps in s.overrides:
                        r0.disabled_sutras.append(so.aps)
                        if s.optional:
                            # Prevent optional sutra's overridden sutras from executing on the same node again
                            v0.disabled_sutras.append(so.aps)
                        logger.debug(f"Disabling overriden {so}")
            # FIXME: disable sutras for AkaqArAdekA saMjYA

            logger.debug(f"O: {r} {[_r.tags for _r in r]} Disabled: {[[s for s in _r.disabled_sutras] for _r in r]}")
            # Update Prakriya Tree
            # Craft inputs and outputs based on viewed inputs
            # And generated outputs
            pnv = node.outputs.copy_replace_at(ix, v[0]).copy_replace_at(ix+1, v[1])
            pnr = node.outputs.copy_replace_at(ix, r[0]).copy_replace_at(ix+1, r[1])
            if len(r) > 2:
                for i in range(len(r)-2):
                    pnr = pnr.copy_insert_at(ix+i+2, r[i+2])
            _ps = PrakriyaNode(pnv, pnr, s, ix, [t for t in triggered if t != s])
            logger.debug(f'O Node: {str(_ps)}')
            if node is not None:
                self.tree.add_child(node, _ps, opt=s.optional)
            else:
                self.tree.add_node(_ps, root=True)
            return r
        else:
            logger.debug(f"Domain {self.domains.active_domain()} - Nothing triggered")
            return False

    def _exec_all_domains(self, node):
        for d in ["saMjYA", "prakfti", "pratyaya", "aNga", "standard", "pada", "saMhitA"]:
            self.domains.set_domain(d)
            r = self._exec_single(node)
            if r:
                return r
        # Only if nothing ever triggered
        return False

    def execute(self):
        if self.need_hier:
            logger.debug(f"Input: {self.hier_inputs}")
        else:
            logger.debug(f"Input: {self.inputs}")
        done = []
        act = False
        # Initial run on input
        for r in self.tree.get_root():
            _act = self._exec_all_domains(r)
            act = act or _act
        if (act):
            # Iterate over leaves if something triggered
            while (act):
                act = False
                for n in self.tree.get_leaves():
                    if n not in done:
                        res = self._exec_all_domains(n)
                        if not res:
                            done.append(n)
                        else:
                            act = True
            for n in self.tree.get_leaves():
                assert n in done
                self.outputs.append(n.outputs)
        else:
            # Nothing triggered
            logger.debug("Nothing Triggered - Passthrough")
            for n in self.tree.get_root():
                self.outputs.append(n.outputs)
        r = self.outputs
        logger.debug(f"Final Result: {r}\n")
        return r

    def output(self, copy=False):
        if copy:
            return deepcopy(self.outputs)
        else:
            return self.outputs

    def describe(self):
        print("\nPrakriya")
        if self.hier_prakriyas != []:
            print(f"Pre Input {self.pre_inputs}")
        for h in self.hier_prakriyas:
            print("Hierarchical Prakriya")
            h.describe()
        print(f"Input {self.inputs}")
        self.tree.describe()
        print(f"Final Output {self.outputs} = {[''.join([str(x) for x in y]) for y in self.outputs]}\n\n")

    def dict(self):
        return self.tree.dict()


_node_id = 0


class PrakriyaNode(object):
    """
    Prakriya History Node

    Inputs
        inputs : list of Paninian Objects
        outputs: list of Paninian Objects
       sutra_id: id for triggered sutra
    other_sutras: sutras that were triggered, but did not win.
    """
    def __init__(self, inputs, outputs, sutra, ix=0, other_sutras=[]):
        global _node_id
        self.id = _node_id
        _node_id = _node_id+1
        self.inputs = inputs
        self.outputs = outputs
        self.sutra = sutra
        self.other_sutras = other_sutras
        self.index = ix

    def __str__(self):
        return f"{self.id} {self.sutra} {self.inputs} {self.index}-> {self.outputs}"

    def __hash__(self):
        return hash("{self.sutra} {self.inputs} {self.index}-> {self.outputs}")

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)

    def describe(self):
        print("Prakriya Node")
        print(str(self))
        if self.other_sutras:
            print("Sutras that were tiggered but did not win")
            for s in self.other_sutras:
                print(str(s))
        print("End")

    def dict(self):
        return {
            'id': self.id,
            'sutra': str(self.sutra),
            'inputs': self.inputs,
            'outputs': self.outputs,
            'window': self.index,
            'other_sutras': [str(s) for s in self.other_sutras]
            }


class PrakriyaTree(object):
    """
    Prakriya Tree: Tree of PrakriyaNodes
    """
    def __init__(self, node=None):
        self.children = {}
        self.parent = {}
        self.leaves = []
        self.roots = []
        if node is not None:
            self.add_node(node)

    def add_node(self, node, root=False):
        self.children[node] = []
        self.leaves.append(node)
        if root:
            self.parent[node] = None
            self.roots.append(node)

    def get_leaves(self):
        return self.leaves

    def get_root(self):
        return self.roots

    def add_child(self, node, c, opt=False):
        self.children[node].append(c)
        assert (c not in self.parent), f"Duplicated {c}, hash{c}"
        self.parent[c] = node
        if c not in self.children:
            self.add_node(c)
        if (not opt) and (node in self.leaves):
            self.leaves.remove(node)

    def describe(self):
        def _desc(n):
            n.describe()
            if n in self.leaves:
                print("Leaf Node")
            for c in self.children[n]:
                print("Child")
                _desc(c)
        for r in self.roots:
            print("Root")
            _desc(r)

    def dict(self):
        def _dict(n):
            d = n.dict()
            d['children'] = [_dict(c) for c in self.children[n]]
            return d
        return {
            'root': _dict(self.roots[0])
            }


def _isScalar(x):
    # We do not expect np arrays or other funky nonscalars here
    return not (isinstance(x, list) or isinstance(x, tuple))


if __name__ == "__main__":
    from sanskrit_parser.generator.pratyaya import *   # noqa: F401, F403
    from sanskrit_parser.generator.dhatu import *   # noqa: F401, F403
    from sanskrit_parser import enable_console_logger
    from sandhi_yaml import sutra_list
    from argparse import ArgumentParser

    def getArgs():
        """
          Argparse routine.
          Returns args variable
        """
        # Parser Setup
        parser = ArgumentParser(description='Prakriya Utility')
        # Input Encoding (autodetect by default)
        parser.add_argument('inputs', type=str, nargs="*",
                            default=None)
        return parser.parse_args()

    def main(args):
        enable_console_logger()
        p = Prakriya(sutra_list, PrakriyaVakya(args.inputs))
        p.execute()
        p.describe

    main(getArgs())



"""
Paninian Object Class

Derived from SanskritObject

@author: kmadathil

"""
from sanskrit_parser.base.sanskrit_base import SanskritObject, SLP1
import logging
logger = logging.getLogger(__name__)


class PaninianObject(SanskritObject):
    """ Paninian Object Class: Derived From SanskritObject

    Attributes:
    """
    def __init__(self, thing=None, encoding=None, unicode_encoding='utf-8',
                 strict_io=True, replace_ending_visarga='s'):
        super().__init__(thing, encoding, unicode_encoding, strict_io,
                         replace_ending_visarga)
        self.inPrakriya = True
        # FIXME: I don't like this being here
        self.disabled_sutras = []
        # Prakriya Related Tags are ephemeral

    def hasTag(self, t):
        return t in self.tags

    def deleteTag(self, t):
        return self.tags.remove(t)

    def setTag(self, t):
        if t not in self.tags:
            self.tags.append(t)
        return t

    def fix(self):
        self.inPrakriya = False

    def isPada(self):
        return self.hasTag("pada")

    @classmethod
    def join_objects(cls, objects):
        logger.debug(f"Joining Objects {objects} {type(objects)}")
        for o in objects[0]:
            logger.debug(f"{o} type {type(o)}")
            assert isinstance(o, SanskritObject), f"{o} type {type(o)}"
        s = "".join([o.canonical() for o in objects[0]])
        so = PaninianObject(s, encoding=SLP1)
        # Tag rules
        # 1.4.14 suptiNantaM padam
        if objects[0][-1].hasTag("sup") or objects[0][-1].hasTag("tiN"):
            so.setTag("pada")
        # 1.4.13 yasmAtpratyayaviDistadAdipratyayeNgam
        elif objects[0][0].hasTag("aNga"):
            so.setTag("aNga")
        # 3.1.32 sannAdyantA dhAtavaH
        if objects[0][-1].hasTag("sannAdi"):
            so.setTag("DAtu")
        # 1.2.46 krttaDitasamAsAsca
        if objects[0][-1].hasTag("krt") or objects[0][-1].hasTag("tadDita"):
            so.setTag("prAtipadika")

        # Custom tag propagation for rule implementation
        for t in ["eti", "eDati", "UW", "sTA", "sTamB"]:
            if objects[0][0].hasTag(t) and objects[0][0].hasTag("DAtu"):
                so.setTag(t)
        for t in ["AN"]:
            if objects[0][0].hasTag(t) and objects[0][0].hasTag("upasarga"):
                so.setTag(t)
        for t in ["trc", "trn"]:
            if objects[0][-1].hasTag(t) and objects[0][0].hasTag("aNga"):
                so.setTag(t)
        for t in ["NI", "Ap", 'strI_abs']:
            if objects[0][-1].hasTag(t):
                so.setTag("strI")
                so.setTag(t)
                for tt in objects[0][0].tags:
                    so.setTag(tt)
                if so.hasTag("pum"):
                    so.deleteTag("pum")
                if so.hasTag("napum"):
                    so.deleteTag("napum")
        return so



from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SLP1
from .operations import adesha
from .maheshvara import ms, isInPratyahara, isSavarna
from sanskrit_parser.generator.pratipadika import *  # noqa: F403


def dirgha(s: str):
    return adesha(s, "aAiIuUfFxX", "AAIIUUFFXX")


def hrasva(s: str):
    return adesha(s, "aAiIuUfFxXeEoO", "aaiiuuffxxiiuu")


def guna(s: str):
    if s in "fF":
        # Uran raparaH
        r = "ar"
    elif s in "xX":
        r = "al"
    else:
        r = adesha(s, "iIuU", "eeoo")
    return r


def vriddhi(s: str):
    if s in "fF":
        # Uran raparaH
        r = "Ar"
    elif s in "xX":
        r = "Al"
    else:
        r = adesha(guna(s), "aeo", "AEO")
    return r


def ikoyan(s: str):
    return adesha(s.lower(),
                  ms.getPratyahara(SanskritImmutableString("ik", SLP1)),
                  ms.getPratyahara(SanskritImmutableString("yaR", SLP1),
                                   remove_a=True)
                  )


def samprasaranam(s: str):
    return adesha(s,
                  ms.getPratyahara(SanskritImmutableString("yaR", SLP1),
                                   remove_a=True),
                  ms.getPratyahara(SanskritImmutableString("ik", SLP1))
                  )


def ayavayav(s: str):
    if s == "e":
        return "ay"
    elif s == "o":
        return "av"
    elif s == "E":
        return "Ay"
    elif s == "O":
        return "Av"
    else:
        return s


def shcutva(s: str):
    return adesha(s, "stTdDn", "ScCjJY")


def zwutva(s: str):
    return adesha(s, "stTdDn", "zwWqQR")


def jashtva(s: str):
    return adesha(s, "JBGQDjbgqdKPCWTcwtkpSzsh", "jbgqdjbgqdgbjqdjqdgbjqdg")


def chartva(s: str):
    return adesha(s, "kKgGcCjJwWqQtTdDpPbB", "kkkkccccwwwwttttpppp")


def kutva(s: str):
    return adesha(s, "cCjJY", "kKgGN")


def vargatritiya(s: str):
    return adesha(s, "kKgGcCjJwWqQtTdDpPbB", "GGGGJJJJQQQQDDDDBBBB")


def adivriddhi(s: str):
    r = ""
    av = False
    for _s in s:
        if (not av) and isInPratyahara("ac", SanskritImmutableString(_s, encoding=SLP1)):
            r = r + vriddhi(_s)
            av = True
        else:
            r = r + _s
    return r


def pAdAdi_AdeSa(s: str):
    d = {
        "pAda": "pad",
        "danta": "dat",
        "nAsikA": "nas",
        "mAsa": "mAs",
        "hfdaya": "hfd",
        "niSA": "niS",
        "asfj": "asan",
        "yUza": "yUzan",
        "doza": "dozan",
        "yakft": "yakan",
        "Sakft": "Sakan",
        "udaka": "udan",
        "Asya": "Asan",
        }
    if s in d:
        return d[s]
    else:
        return s


# FIXME - this is better. debug in SK228
def pAdAdi_AdeSa_r(s):
    d = {
        "pAda": pad,    # noqa: F405
        "danta": dat,    # noqa: F405
        "nAsikA": nas,    # noqa: F405
        "mAsa": mAs,    # noqa: F405
        "hfdaya": hfd,    # noqa: F405
        "niSA": niS,    # noqa: F405
        "asfj": asan,    # noqa: F405
        "yUza": yUzan,    # noqa: F405
        "doza": dozan,    # noqa: F405
        "yakft": yakan,    # noqa: F405
        "Sakft": Sakan,    # noqa: F405
        "udaka": udan,    # noqa: F405
        "Asya": Asan,    # noqa: F405
        }
    if str(s) in d:
        return d[s]
    else:
        return s


# Fixme - anunasika ZSs yrl
def anunasika(s: str):
    if s in "yrlvSZs":
        return s+"~"
    else:
        return adesha(s, "kKgGcCjJwWqQtTdDpPbB", "NNNNYYYYRRRRnnnnmmmm")


# vyavAya check for razAByAM noRaH samAnapade
def rz_vyavaya_l(s: str):
    i = len(s)-1
    while(i >= 0):
        # ऋवर्णात् नस्य णत्वं वाच्यम्
        if ((s[i] == "r") or (s[i] == "z") or (s[i] == "f")):
            return True
        elif awkupvaNnum(s[i]):
            i = i-1
        else:
            return False
    return False


def rz_vyavaya_r(s: str):
    i = 0
    while(i < len(s)):
        if (s[i] == "n"):
            # Na padAntasya
            if s.hasTag("svAdi"):
                return (i != (len(s) - 1))
            else:
                return True
        elif awkupvaNnum(s[i]):
            i = i+1
        else:
            return False
    return False


# For situations like yUzan + i yUzaRi
def rz_vyavaya_n(s: str):
    if len(s) == 0:
        return False
    if (s[-1] != "n"):
        return False
    i = len(s)-2
    while(i >= 0):
        # ऋवर्णात् नस्य णत्वं वाच्यम्
        if ((s[i] == "r") or (s[i] == "z") or (s[i] == "f")):
            return True
        elif awkupvaNnum(s[i]):
            i = i-1
        else:
            return False
    return False


def awkupvaNnum(s):
    # FIXME handle AN
    return isInPratyahara("aw", s) or isSavarna("ku", s) or isSavarna("pu", s) \
        or (s == "M")


def Ratva(s):
    return s.replace("n", "R", 1)


def iyuvaN(s):
    if isSavarna("i", s):
        return "iy"
    elif isSavarna("u", s):
        return ("uv")
    else:
        return s


def anekAc_asaMyogapUrva(s):
    if ((len(s) > 2) and (isInPratyahara("hal", s[-2]) and isInPratyahara("hal", s[-3]))):
        return False
    ac = 0
    for sc in s:
        if isInPratyahara("ac", sc):
            ac = ac+1
        if ac > 1:
            return True
    return False


def numAgama(s):
    lastac = -1
    lens = len(s)
    for j in range(lens):
        jj = -1*(j+1)
        if isInPratyahara("ac", s[jj]):  # Backwards
            lastac = lens + jj
            break
    if lastac == lens-1:
        r = s + "n"
    elif lastac > -1:
        r = s[:lastac+1] + "n" + s[lastac+1:]
    else:
        r = s
    return r


# aco'ntyAdi wi
def wilopa(s):
    lastac = -1
    lens = len(s)
    for j in range(lens):
        jj = -1*(j+1)
        if isInPratyahara("ac", s[jj]):  # Backwards
            lastac = lens + jj
            break
    if lastac > -1:
        r = s[:lastac]
    else:
        r = s
    return r


def notnull(s):
    return ((s is not None) and (s != ""))


def null(s):
    return ((s is None) or (s == ""))


# sUtra: adeN guRaH
def is_guna(s: str):
    so = SanskritImmutableString(s, encoding=SLP1)
    at = "a"
    eng = SanskritImmutableString("eN", encoding=SLP1)
    return (s == at) or ms.isInPratyahara(eng, so)


# sUtra: vRdDirAdEc
def is_vriddhi(s: str):
    so = SanskritImmutableString(s, encoding=SLP1)
    aat = "A"
    aich = SanskritImmutableString("Ec", encoding=SLP1)
    return (s == aat) or ms.isInPratyahara(aich, so)



from sanskrit_parser.generator.paninian_object import PaninianObject
from sanskrit_parser.base.sanskrit_base import SLP1


class Pratyaya(PaninianObject):
    """ Sanskrit Object Class: Derived From SanskritString

     Attributes:
    """
    def __init__(self, thing=None, its=[], other_tags=[], encoding=SLP1,
                 unicode_encoding='utf-8',
                 strict_io=True, replace_ending_visarga='s'):
        super().__init__(thing, encoding, unicode_encoding, strict_io, replace_ending_visarga)
        self.inPrakriya = True
        self.its = its
        self.setTag("pratyaya")
        for t in other_tags:
            self.setTag(t)

    def hasIt(self, it):
        return it in self.its

    def setIt(self, it):
        if it not in self.its:
            self.its.append(it)

    def deleteIt(self, it):
        return self.its.remove(it)

    def luTags(self):
        _l = []
        for t in self.tags:
            if not (t in ["luk", "Slu", "lup"]):
                _l.append(t)
        for t in _l:
            self.tags.remove(t)


tuk = Pratyaya("t", its=["k"])

# Aw (agama)
Aw = Pratyaya("A", its=["w"], other_tags=["Aw"])
yAw = Pratyaya("yA", its=["w"], other_tags=["yAw"])
syAw = Pratyaya("syA", its=["w"], other_tags=["syAw"])

tip = Pratyaya("ti", its=["p"], other_tags=["sArvaDAtuka"])
Sap = Pratyaya("a", its=["S", "p"], other_tags=["sArvaDAtuka"])


Ryat = Pratyaya("ya", its=["t", "R"], other_tags=["ArDaDAtuka"])
GaY = Pratyaya("a", its=["G", "Y"], other_tags=["ArDaDAtuka"])
Ric = Pratyaya("i", its=["R", "c"], other_tags=["ArDaDAtuka"])
tfc = Pratyaya("tf", its=["c"], other_tags=["ArDaDAtuka"])
yat = Pratyaya("ya", its=["t"], other_tags=["ArDaDAtuka"])
yak = Pratyaya("ya", its=["k"], other_tags=["ArDaDAtuka"])
ktvA = Pratyaya("tvA", its=["k"], other_tags=["ArDaDAtuka", "avyaya"])
kta = Pratyaya("ta", its=["k"], other_tags=["ArDaDAtuka"])

# tiN
tip = Pratyaya("ti", its=["p"], other_tags=["tiN", "sArvaDAtuka"])
sip = Pratyaya("si", its=["p"], other_tags=["tiN", "sArvaDAtuka"])


# nIpAtAs
AN = Pratyaya("A", its=["N"], other_tags=["nipAta", "upasarga", "pada"])
mAN = Pratyaya("mA", its=["N"], other_tags=["nipAta", "upasarga", "pada"])
upa = Pratyaya("upa", other_tags=["nipAta", "upasarga", "pada"])
pra = Pratyaya("pra", other_tags=["nipAta", "upasarga", "pada"])
ava = Pratyaya("ava", other_tags=["nipAta", "upasarga", "pada"])
ud = Pratyaya("ud", other_tags=["nipAta", "upasarga", "pada"])


# bha when applied to prAtipadikas only!
# FIXME - maybe have two yats?
yat_t = Pratyaya("ya", its=["t"], other_tags=["svAdi", "tadDita"])
zyaY_t = Pratyaya("ya", its=["z", "Y"], other_tags=["svAdi", "tadDita"])
yaY_t = Pratyaya("ya", its=["Y"], other_tags=["svAdi", "tadDita"])
aR_t = Pratyaya("a", its=["R"], other_tags=["svAdi", "tadDita"])

# sup
# स्वौजसमौट्छष्टाभ्याम्भिस्ङेभ्याम्भ्यस्ङसिभ्याम्भ्यस्ङसोसाम्ङ्योस्सुप्
su = Pratyaya("s", its=["u~"], other_tags=["svAdi", "sup",  "su", "suw"])
O = Pratyaya("O", other_tags=["svAdi", "sup", "O", "suw"])   # noqa 741
jas = Pratyaya("as", its=["j"], other_tags=["svAdi", "sup", "jas",
                                            "suw"])
am = Pratyaya("am", other_tags=["svAdi", "sup", "suw", "am"])
Ow = Pratyaya("O", its=["w"], other_tags=["svAdi", "Ow", "sup", "suw"])
Sas = Pratyaya("as", its=["S"], other_tags=["svAdi", "sup", "Sas"])
wA = Pratyaya("A", its=["w"], other_tags=["svAdi", "wA", "sup"])
ByAm = Pratyaya("ByAm", other_tags=["svAdi", "ByAm", "sup"])
Bis = Pratyaya("Bis", other_tags=["svAdi", "Bis", "sup"])
Ne = Pratyaya("e", its=["N"], other_tags=["svAdi", "Ne", "sup"])
ByAm2 = Pratyaya("ByAm", other_tags=["svAdi", "ByAm", "sup"])
Byas = Pratyaya("Byas", other_tags=["svAdi", "Byas", "sup"])
Nasi = Pratyaya("as", its=["N", "i~"], other_tags=["svAdi", "Nasi", "sup"])
ByAm3 = Pratyaya("ByAm", other_tags=["svAdi", "ByAm", "sup"])
Byas2 = Pratyaya("Byas", other_tags=["svAdi",  "Byas", "sup"])
Nas = Pratyaya("as", its=["N"], other_tags=["svAdi", "Nas", "sup"])
os = Pratyaya("os", other_tags=["svAdi", "os", "sup"])
Am = Pratyaya("Am", other_tags=["svAdi", "Am", "sup"])
Ni = Pratyaya("i", its=["N"], other_tags=["svAdi", "Ni", "sup"])
os2 = Pratyaya("os", other_tags=["svAdi", "os", "sup"])
sup = Pratyaya("su", its=["p"], other_tags=["svAdi", "sup"])
su2 = Pratyaya("s", its=["u~"], other_tags=["svAdi", "sup", "suw", "su",
                                            "sambudDi"])
O2 = Pratyaya("O", other_tags=["svAdi", "O", "sup", "suw"])
jas2 = Pratyaya("as", its=["j"], other_tags=["svAdi", "sup", "suw", "jas"])

sups = [[su, O, jas],
        [am, Ow, Sas],
        [wA, ByAm, Bis],
        [Ne, ByAm2, Byas],
        [Nasi, ByAm3, Byas2],
        [Nas, os, Am],
        [Ni, os, sup],
        [su2, O2, jas2]]

for p in sups:
    for ix, v in enumerate(["ekavacana", "dvivacana", "bahuvacana"]):
        p[ix].setTag(v)

for ix, v in enumerate(["praTamA", "dvitIyA", "tftIyA", "caturTi",
                        "pancamI", "zazWI", "saptamI", "samboDana"]):
    for p in sups[ix][:]:
        p.setTag(v)
        p.setTag("viBakti")

# SI - jasaH SI
SI = Pratyaya("I", its=["S"], other_tags=["svAdi", "sup", "SI"])
# Si - jasSasaH Si
Si = Pratyaya("i", its=["S"], other_tags=["svAdi", "sup", "Si",
                                          "sarvanAmasTAna"])
# adaw - for watarAdi
adaq = Pratyaya("ad", its=["q"], other_tags=["svAdi", "sup", "adaq"])

# StrI
NIp = Pratyaya("I", its=["N", "p"], other_tags=["svAdi", "NI", "strI"])
NIz = Pratyaya("I", its=["N", "z"], other_tags=["svAdi", "NI", "strI"])
Ap = Pratyaya("A", its=["p"], other_tags=["svAdi", "Ap", "strI"])
strI_abs = Pratyaya("", its=[], other_tags=["strI_abs", "strI"])

# Sup Luk
luk_sup = Pratyaya("", its=[], other_tags=["sup"])

# Sambuddhi
sambudDi = PaninianObject("")
sambudDi.setTag("sambudDi")

# Anta
avasAna = PaninianObject(".")
avasAna.setTag("avasAna")



from sanskrit_parser.base.maheshvara_sutra import MaheshvaraSutras
from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SLP1

ms = MaheshvaraSutras()


# Utility functions
# Pratyahara check
def isInPratyahara(p, s):
    if s == "":
        return False
    if isinstance(p, str):
        p = SanskritImmutableString(p, SLP1)
    if isinstance(s, str):
        s = SanskritImmutableString(s, SLP1)
    return ms.isInPratyahara(p, s)


# Savarna check
def isSavarna(p, s):
    if ((s == "") and (p != "") or (s != "") and (p == "")):
        return False
    elif ((s == "") and (p == "")):
        return True
    else:
        if isinstance(p, str):
            p = SanskritImmutableString(p, SLP1)
        if isinstance(s, str):
            s = SanskritImmutableString(s, SLP1)
        return ms.isSavarna(p, s)


__all__ = ["isInPratyahara", "isSavarna"]



# -*- coding: utf-8 -*-
"""

Usage
======

The ``Parser`` class can be used to generate vakya parses thus:

.. code-block:: python

        from itertools import islice
        from sanskrit_parser import Parser
        string = "astyuttarasyAMdiSi"
        input_encoding = "SLP1"
        output_encoding = "SLP1"
        parser = Parser(input_encoding=input_encoding,
                        output_encoding=output_encoding,
                        replace_ending_visarga='s')
        print('Splits:')
        for split in parser.split(string, limit=10):
            print(f'Lexical Split: {split}')
            for i, parse in enumerate(split.parse(limit=2)):
                print(f'Parse {i}')
                print(f'{parse}')
            break


This produces the output::

    Lexical Split: ['asti', 'uttarasyAm', 'diSi']
    Parse 0
    asti => (asti, ['samAsapUrvapadanAmapadam', 'strIliNgam']) : samasta of uttarasyAm
    uttarasyAm => (uttara#1, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam'])
    diSi => (diS, ['saptamIviBaktiH', 'ekavacanam', 'strIliNgam']) : viSezaRa of uttarasyAm
    Parse 1
    asti => (asti, ['samAsapUrvapadanAmapadam', 'strIliNgam']) : samasta of uttarasyAm
    uttarasyAm => (uttara#2, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam']) : viSezaRa of diSi
    diSi => (diS#2, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam'])
    Parse 2
    asti => (as#1, ['kartari', 'praTamapuruzaH', 'law', 'parasmEpadam', 'ekavacanam', 'prATamikaH'])
    uttarasyAm => (uttara#2, ['saptamIviBaktiH', 'strIliNgam', 'ekavacanam']) : viSezaRa of diSi
    diSi => (diS, ['saptamIviBaktiH', 'ekavacanam', 'strIliNgam']) : aDikaraRam of asti

@author: avinashvarna
"""

import time
import json
import abc
import warnings
from dataclasses import dataclass
from typing import Sequence
import itertools
from sanskrit_parser.base.sanskrit_base import SCHEMES, SanskritObject, SLP1
from sanskrit_parser.base.sanskrit_base import SanskritNormalizedString, SanskritString
from sanskrit_parser.parser.sandhi_analyzer import LexicalSandhiAnalyzer
from sanskrit_parser.parser.datastructures import VakyaGraph, VakyaGraphNode
import logging

logger = logging.getLogger(__name__)


class Serializable(abc.ABC):
    """ Base class to indicate an object is serializable into JSON """

    @abc.abstractmethod
    def serializable(self):
        ''' Return an object that can be serialized by json.JSONEncoder '''
        pass


class JSONEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, Serializable):
            return o.serializable()
        return super().default(o)


class Parser():

    def __init__(self, strict_io: bool = False, input_encoding: str = None,
                 output_encoding: str = 'SLP1', lexical_lookup: str = "combined",
                 score: bool = True, split_above: int = 5,
                 replace_ending_visarga: str = None, fast_merge: bool = True):
        self.strict_io = strict_io
        if input_encoding is not None:
            self.input_encoding = SCHEMES[input_encoding]
        else:
            self.input_encoding = None
        self.output_encoding = SCHEMES[output_encoding]
        self.lexical_lookup = lexical_lookup
        self.score = score
        self.split_above = split_above
        self.replace_ending_visarga = replace_ending_visarga
        self.fast_merge = fast_merge
        self.sandhi_analyzer = LexicalSandhiAnalyzer(self.lexical_lookup)

    def _maybe_pre_segment(self, input_string: str, pre_segmented: bool
                           ):
        ''' Pre-process pre-segmented input if necessary '''
        if not pre_segmented:
            s = SanskritNormalizedString(input_string,
                                         encoding=self.input_encoding,
                                         strict_io=self.strict_io,
                                         replace_ending_visarga=self.replace_ending_visarga)
            logger.info(f"Input String in SLP1: {s.canonical()}")
            return s
        else:
            logger.debug("Pre-Segmented")
            s = []
            for seg in input_string.split(" "):
                o = SanskritObject(seg,
                                   encoding=self.input_encoding,
                                   strict_io=self.strict_io,
                                   replace_ending_visarga='r')
                ts = self.sandhi_analyzer.getMorphologicalTags(o, tmap=True)
                if ts is None:
                    # Possible sakaranta
                    # Try by replacing end visarga with 's' instead
                    o = SanskritObject(seg,
                                       encoding=self.input_encoding,
                                       strict_io=self.strict_io,
                                       replace_ending_visarga='s')
                    ts = self.sandhi_analyzer.getMorphologicalTags(o, tmap=True)
                if ts is None:
                    logger.warning(f"Unknown pada {seg} - will be split")
                    _p = self.parse(seg, False)
                    _s = list(itertools.islice(_p.splits(), 1))[0]
                    logger.info(f"Split {_s}")
                    s.extend(_s.split)
                    if _s is None:
                        logger.warning(f"Unknown pada {seg} - cannot be split")
                else:
                    s.append(o)
            logger.info(f"Input String in SLP1: {' '.join([x.canonical() for x in s])}")
            return s

    def split(self,
              input_string: str,
              limit: int = 10,
              pre_segmented: bool = False,
              ):
        s = self._maybe_pre_segment(input_string, pre_segmented)
        logger.debug("Start Split")
        graph = self.sandhi_analyzer.getSandhiSplits(s, tag=True,
                                                     pre_segmented=pre_segmented)
        logger.debug("End DAG generation")
        if graph is None:
            warnings.warn("No splits found. Please check the input to ensure there are no typos.")
            return None
        else:
            splits = graph.find_all_paths(max_paths=limit,
                                          sort=True,
                                          score=self.score)
            return [Split(self, input_string, split)
                    for split in splits]


@dataclass
class Split(Serializable):
    parser: Parser
    input_string: str
    split: Sequence[SanskritObject]
    vgraph: VakyaGraph = None

    def __repr__(self):
        return f'Split({self.input_string}) = {self.split}'

    def __str__(self):
        strict_io = self.parser.strict_io
        encoding = self.parser.output_encoding
        out = [t.transcoded(encoding, strict_io) for t in self.split]
        return str(out)

    def parse(self, limit=10, min_cost_only=False):
        self.vgraph = VakyaGraph(self.split,
                                 fast_merge=self.parser.fast_merge,
                                 max_parse_dc=self.parser.split_above)
        parses = self.vgraph.parses[:limit]
        costs = self.vgraph.parse_costs[:limit]
        min_cost = costs[0]
        if min_cost_only:
            parses = [x for x, cost in zip(parses, costs) if cost == min_cost]
        return [Parse(self, parse, cost) for parse, cost in zip(parses, costs)]

    def write_dot(self, basepath):
        self.vgraph.write_dot(basepath)

    def serializable(self):
        strict_io = self.parser.strict_io
        encoding = self.parser.output_encoding
        out = [t.transcoded(encoding, strict_io) for t in self.split]
        return {'split': out,
                'parses': list(self.parse())
                }


@dataclass
class ParseTag(Serializable):
    root: str
    tags: Sequence[str]

    def serializable(self):
        return {'root': self.root, 'tags': self.tags}

    def __str__(self):
        return f'({self.root}, {self.tags})'


class ParseNode(Serializable):
    def __init__(self, node: VakyaGraphNode,
                 strict_io: bool,
                 encoding: str):
        self.pada = node.pada.transcoded(encoding, strict_io)
        tag = node.getMorphologicalTags()
        self.parse_tag = ParseTag(tag[0].transcoded(encoding, strict_io),
                                  [t.transcoded(encoding, strict_io) for t in tag[1]]
                                  )
        self.index = node.index

    def __str__(self):
        return f'{self.pada} => {self.parse_tag}'

    def serializable(self):
        d = {'pada': self.pada}
        d.update(self.parse_tag.serializable())
        return d


@dataclass
class ParseEdge(Serializable):
    predecessor: ParseNode
    node: ParseNode
    label: str

    def __str__(self):
        return f'{self.node} : {self.label} of {self.predecessor.pada}'

    def serializable(self):
        return {'node': self.node,
                'predecessor': self.predecessor,
                'sambandha': self.label
                }


class Parse(Serializable):

    def __init__(self, split: Split, parse_graph, cost):
        strict_io = split.parser.strict_io
        encoding = split.parser.output_encoding
        graph = []
        for n in sorted(list(parse_graph), key=lambda x: x.index):
            node = ParseNode(n, strict_io, encoding)
            preds = list(parse_graph.predecessors(n))
            if preds:
                pred = preds[0]  # Only one
                # Multigraph
                # Ugh - surely this can be better
                lbl = list(parse_graph[pred][n].values())[0]['label']
                pred_node = ParseNode(pred, strict_io, encoding)
                edge = ParseEdge(pred_node,
                                 node,
                                 SanskritString(lbl, encoding=SLP1).transcoded(encoding, strict_io)
                                 )
                graph.append(edge)
            else:
                graph.append(node)
        self.graph = graph
        self.cost = cost
        self.parse_graph = parse_graph

    def __str__(self):
        return '\n'.join([str(t) for t in self.graph])

    def __iter__(self):
        return iter(self.graph)

    def to_conll(self):
        r = []
        for t in self.graph:
            if isinstance(t, ParseNode):
                r.append([str(t.index+1), str(t.pada), "_",
                          str(t.parse_tag.root), str(t.parse_tag.tags),
                          "0", "root"])
            else:
                r.append([str(t.node.index+1), str(t.node.pada), "_",
                          str(t.node.parse_tag.root), str(t.node.parse_tag.tags),
                          str(t.predecessor.index+1), str(t.label)])
        return r

    def to_dot(self):
        from io import StringIO
        import networkx as nx
        s = StringIO()
        nx.drawing.nx_pydot.write_dot(self.parse_graph, s)
        return s.getvalue()

    def serializable(self):
        return {'graph': self.graph}


if __name__ == "__main__":
    start_time = time.time()

    def api_example(string, output_encoding):
        parser = Parser(output_encoding=output_encoding,
                        replace_ending_visarga='s')
        print('Splits:')
        for split in parser.split(string, limit=10):
            print(f'Lexical Split: {split}')
            for i, parse in enumerate(split.parse(limit=2)):
                print(f'Parse {i}')
                print(f'{parse}')
            break
        print(json.dumps(split,
                         ensure_ascii=False,
                         indent=2,
                         cls=JSONEncoder)
              )

    def main():
        examples = [('devadattogrAmaMgacCati', 'SLP1'),
                    ('astyuttarasyAMdishidevatAtmA', 'Devanagari')
                    ]
        for string, encoding in examples:
            api_example(string, encoding)

    main()
    print(f'Took {time.time() - start_time} s')



"""
Operational Sutras

"""
from sanskrit_parser.base.sanskrit_base import SanskritImmutableString, SLP1
from decimal import Decimal
from copy import deepcopy
from sanskrit_parser.generator.paninian_object import PaninianObject

import logging
logger = logging.getLogger(__name__)


# Global Domains
class GlobalDomains(object):
    def __init__(self):
        self.domains = {
            "saMjYA": True,
            "upadeSa": False,
            "prakfti": False,
            "pratyaya": False,
            "aNga": False,
            "pada": False,
            "saMhitA": False,
            "standard": False
        }

    def isdomain(self, d):
        return self.domains[d]

    def set_domain(self, d):
        for k in self.domains:
            if k == d:
                self.domains[k] = True
            else:
                self.domains[k] = False

    def active_domain(self):
        r = []
        for k in self.domains:
            if self.domains[k]:
                r.append(k)
        return r


# Base class
class Sutra(object):
    def __init__(self, name, aps, optional=False, overrides=None):
        if isinstance(name, str):
            self.name = SanskritImmutableString(name)
        else:
            self.name = name
        if isinstance(aps, str):
            self.aps = aps   # Adhaya.pada.sutra
            aps_l = aps.split(".")
            aps_t = [int(_x) for _x in aps_l]
            if len(aps_l) > 3:  # Subsutra/Vartikam
                aps_sub = Decimal("0."+str(aps_t[-1]))
            else:
                aps_sub = 0
            self._aps_tuple = aps_t
        elif isinstance(aps, tuple):
            aps_t = aps
            self._aps_tuple = aps_t
            self.aps = '.'.join([str(x) for x in list(aps_t)])
        self._aps_num = aps_t[2]+aps_t[1]*1000+aps_t[0]*10000 + aps_sub
        self.overrides = overrides
        self.optional = optional
        logger.info(f"Initialized {self}:  {self._aps_num} Optional:{self.optional}")

    def __str__(self):
        if self.optional:
            _o = "*"
        else:
            _o = ""
        return f"{self.aps:7}: {str(self.name)} {_o}"


class LRSutra(Sutra):
    def __init__(self, name, aps, cond, xform, insert=None, domain=None,
                 update=None, optional=False, bahiranga=1, overrides=None):
        '''
        Sutra Class that expects a left and right input
        '''
        super().__init__(name, aps, optional, overrides)
        self.domain = domain
        self.cond = cond
        self.xform = xform
        self.update_f = update
        self.insertx = insert
        self.bahiranga = bahiranga  # Bahiranga score. Smaller wins

    def inAdhikara(self, context):
        return self.adhikara(context)

    def isTriggered(self, s1, s2, domains):
        logger.debug(f"Checking {self} View: {s1} {s2}")
        env = _env(s1, s2)
        if self.domain is not None:
            t = self.domain(domains)
        else:
            t = domains.isdomain("standard")
        if self.cond is not None:
            c = self.cond(env)
        else:
            c = True
        logger.debug(f"Check Result {c and t} for {self}")
        return c and t

    def update(self, s1, s2, o1, o2, domains):
        env = _env(s1, s2)
        env["olp"] = o1
        env["orp"] = o2
        if self.update_f is not None:
            self.update_f(env, domains)
        return env["olp"], env["orp"]

    def operate(self, s1, s2):
        # We take the string tuple returned, and update s1, s2
        rs1 = deepcopy(s1)
        rs2 = deepcopy(s2)
        if self.xform is not None:
            env = _env(s1, s2)
            ret = self.xform(env)
            rs1.update(ret[0], SLP1)
            rs2.update(ret[1], SLP1)
        return rs1, rs2

    def insert(self, s1, s2):
        if self.insertx is not None:
            env = _env(s1, s2)
            itx = self.insertx(env)
            r = [s1, s2]
            for i in itx:
                if not isinstance(itx[i], PaninianObject):
                    assert isinstance(itx[i], str)
                    itx[i] = PaninianObject(itx[i])
                r.insert(i, itx[i])
            logger.debug(f"After insertion {r}")
            return r
        else:
            return(s1, s2)


def _env(s1, s2):
    # Helper function to define execution environment
    env = {}
    env["lp"] = s1
    env["rp"] = s2
    if s1.canonical() == "":
        env["l"] = SanskritImmutableString("")
    else:
        env["l"] = SanskritImmutableString(s1.canonical()[-1], SLP1)
    if s2.canonical() == "":
        env["r"] = SanskritImmutableString("")
    else:
        env["r"] = SanskritImmutableString(s2.canonical()[0], SLP1)
    if len(s1.canonical()) > 1:
        env["ll"] = SanskritImmutableString(s1.canonical()[-2], SLP1)
        env["lc"] = SanskritImmutableString(s1.canonical()[:-1], SLP1)
    else:
        env["ll"] = SanskritImmutableString("")
        env["lc"] = SanskritImmutableString("")
    if len(s2.canonical()) > 1:
        env["rr"] = SanskritImmutableString(s2.canonical()[1], SLP1)
        env["rc"] = SanskritImmutableString(s2.canonical()[1:], SLP1)
    else:
        env["rr"] = SanskritImmutableString("", SLP1)
        env["rc"] = SanskritImmutableString("", SLP1)
    return env



# -*- coding: utf-8 -*-
"""
ac sandhi sutras generated through YAML processing

@author: kmadathil
"""
from sanskrit_parser.generator.process_yaml import process_yaml
import yaml
import os.path

import logging
logger = logging.getLogger(__name__)

sutra_list = []
sutra_dict = {}

f = open(os.path.join(os.path.dirname(__file__), "sutras.yaml"), "r")
y = yaml.load(f, Loader=yaml.FullLoader)
sutra_dict = process_yaml(y)
sutra_list = sutra_dict.values()
# logger.info(sutra_dict)
f.close()

__all__ = ["sutra_list", "sutra_dict"]



# -*- coding: utf-8 -*-
"""
Intro
=====
Experimental

Sanskrit pada generator

Process
1. PrakRiti (dhatu / prAtipadika)
2. pratyaya addition
  2.a strIpratyaya
  2.b samAsa
3. prakRti / pratyaya transformations
4. aNga transformations
5. prakRti + pratyaya sandhi
6. pada sandhi

@author: Karthik Madathil (github: @kmadathil)
"""



# -*- coding: utf-8 -*-
"""
Sutra YAML Processor

@author: kmadathil
"""
from sanskrit_parser.generator.sutra import LRSutra
from sanskrit_parser.generator.maheshvara import *  # noqa: F403
from sanskrit_parser.generator.paribhasha import *  # noqa: F403
from sanskrit_parser.generator.pratyaya import *  # noqa: F403
from sanskrit_parser.generator.pratipadika import *  # noqa: F403

import logging
logger = logging.getLogger(__name__)


def process_yaml(y):
    '''
    Process yaml file to return sutras dict

    Inputs:
      y: Sutra Yaml
    Outputs
      sutra_dict: dict of sutras keyed by sutra id
    '''
    logger.debug(f'Processing YAML {y}')
    sutra_dict = {}
    for s in y:
        logger.debug(f"processing {s}")
        if "sutra" not in s:
            logger.error("No sutra name")
            assert False
        if "id" not in s:
            logger.error("No sutra id")
            assert False
        for c in ["condition", "domain", "xform", "update", "insert"]:
            if c not in s:
                s[c] = None
        if "bahiranga" not in s:
            s["bahiranga"] = 1
        # svar = "sutra_"+s["id"].replace(".", "_")
        sname = s["sutra"]
        soverrides = None
        sopt = False
        if "optional" in s:
            sopt = s["optional"]
        if "overrides" in s:
            if isinstance(s["overrides"], str):
                soverrides = [s["overrides"]]
            else:
                soverrides = s["overrides"]
            logger.debug(f"Sutra {s['id']} Overrides {soverrides}")
        scond = None
        if s["condition"] is not None:
            logger.debug("Processing Condition")

            def _exec_cond(s):
                logger.debug(f"Cond dict {s}")
                # FIXME Fix variables after fixing sutra_engine

                def _cond(env):
                    def _c(env):
                        # _s a dict
                        # LHS = variable
                        # RHS = _pratyahara , {variable}, or savarna
                        x = True
                        for k in _s:
                            logger.debug(f"Checking cond {_s[k]} against {k}")

                            def _cond_single(sk, k):
                                if (sk[0] == "_"):
                                    # Pratyahara
                                    logger.debug(f"Checking pratyahara {sk[1:]} {k}")
                                    _x = isInPratyahara(sk[1:], k)  # noqa: F405
                                elif (sk[0:2] == "$$"):
                                    # function call
                                    logger.debug(f"Checking function {sk[2:]} {k}")
                                    _x = eval(f"{sk[2:]}(k)")
                                elif (sk[0] == "$"):
                                    # Variable
                                    logger.debug(f"Checking variable {sk[1:]} {k}")
                                    _x = isSavarna(env[sk[1:]], k)  # noqa: F405
                                elif (sk[0] == "="):
                                    # Raw equality
                                    logger.debug(f"Checking raw {sk[1:]} {k}")
                                    _x = (sk[1:] == k.canonical())
                                elif (sk[0:2] == "!="):
                                    # Raw inequality
                                    logger.debug(f"Checking raw inequality {sk[2:]} {k}")
                                    _x = (sk[2:] != k.canonical())
                                elif (sk[0:2] == "?!"):  # Tag false check
                                    logger.debug(f"Checking tag false {sk[2:]} {k}")
                                    _x = not k.hasTag(sk[2:])
                                elif (sk[0] == "?"):  # Tag check
                                    logger.debug(f"Checking tag {sk[1:]} {k}")
                                    _x = k.hasTag(sk[1:])
                                elif (sk[0] == "+"):  # It check
                                    logger.debug(f"Checking it {sk[1:]} {k}")
                                    _x = k.hasTag("pratyaya") and k.hasIt(sk[1:])
                                else:
                                    logger.debug(f"Checking savarna {sk} {k} ")
                                    _x = isSavarna(sk, k)   # noqa: F405
                                logger.debug(f"Return {_x}")
                                return _x
                            if isinstance(_s[k], list):
                                logger.debug("List")
                                if _s[k][0] == "and":
                                    logger.debug("Checking and condition")
                                    _x = True
                                    for sk in _s[k][1:]:
                                        _x = _x and _cond_single(sk, env[k])
                                else:
                                    _x = False
                                    for sk in _s[k]:
                                        _x = _x or _cond_single(sk, env[k])
                            else:
                                logger.debug("Single")
                                _x = _cond_single(_s[k], env[k])
                            logger.debug(f"Got {_x}")
                            x = x and _x
                        return x
                    # List implies an or condition
                    if isinstance(s, list):
                        _ret = False
                        for _s in s:
                            _ret = _ret or _c(env)
                        return _ret
                    else:
                        _s = s
                        return _c(env)
                return _cond

            scond = _exec_cond(s["condition"])

        sxform = None
        if s["xform"] is not None:
            logger.debug("Processing Xform")

            def _exec_xform(s):
                logger.debug(f"Xform dict {s}")
                xdict = s
                # FIXME Fix variables after fixing sutra_engine

                def _xform(env):
                    # Don't Remove - keep for debug
                    # logger.debug(f"Env {env}")
                    # for k in env:
                    #    logger.debug(f"{k} {type(env[k])}")
                    _l = l = env["l"].canonical()  # noqa: E741, F841
                    _r = r = env["r"].canonical()  # noqa: E741, F841
                    _lc = lc = env["lc"].canonical()  # noqa: E741, F841
                    _rc = rc = env["rc"].canonical()  # noqa: E741, F841
                    logger.debug(f"Xform dict {xdict}")
                    logger.debug(f"Before: {_lc} {_l} {_r} {_rc}")
                    # Execute transforms for predefined variables
                    # FIXME: We assume our code in xform is safe to eval
                    if "l" in xdict:
                        if xdict["l"] is not None:
                            _l = eval(xdict["l"])
                        else:
                            _l = ""
                    if "r" in xdict:
                        if xdict["r"] is not None:
                            _r = eval(xdict["r"])
                        else:
                            _r = ""
                    if "lc" in xdict:
                        if xdict["lc"] is not None:
                            _lc = eval(xdict["lc"])
                        else:
                            _lc = ""
                    if "rc" in xdict:
                        if xdict["rc"] is not None:
                            _rc = eval(xdict["rc"])
                        else:
                            _rc = ""
                    logger.debug(f"After {_lc} {_l} {_r} {_rc}")
                    ret = [_lc+_l, _r+_rc]
                    return ret
                return _xform
            sxform = _exec_xform(s["xform"])
            logger.debug(f"Xform def {sxform}")
        sinsert = None
        if s["insert"] is not None:
            logger.debug("Processing insert")

            def _exec_insert(s):
                logger.debug(f"insert dict {s}")
                idict = s

                def _insert(env):
                    _r = {}
                    for i in idict:
                        logger.debug(f"Insert {i} {idict[i]}")
                        _r[i] = eval(idict[i])
                    return _r
                return _insert
            sinsert = _exec_insert(s["insert"])
            logger.debug(f"Insert def {sinsert}")
        sdom = None
        if s["domain"] is not None:
            logger.debug("Processing domain")

            def _exec_trig(s):
                logger.debug(f"Trig {s}")

                def _trig(domains):
                    # list of domains
                    logger.debug(f"Domain checks {s}")
                    if isinstance(s, list):
                        x = True
                        for t in s:
                            x = x and domains.isdomain(t)
                    else:
                        return domains.isdomain(s)
                return _trig
            sdom = _exec_trig(s["domain"])
        supdate = None
        if s["update"] is not None:
            logger.debug("Processing update")

            def _exec_update(s):
                logger.debug(f"Update {s}")

                def _update(env, domains):
                    def _c(env):
                        # _s a dict
                        # LHS = variable
                        # RHS = _pratyahara , {variable}, or savarna
                        x = True
                        for k in _s:
                            logger.debug(f"Checking cond {_s[k]} against {k}")
                            if (_s[k][0] == "_"):
                                # Pratyahara
                                logger.debug(f"Checking pratyahara {_s[k][1:]}")
                                _x = isInPratyahara(_s[k][1:], env[k])  # noqa: F405
                            elif (_s[k][0] == "$"):
                                # Variable
                                logger.debug(f"Checking variable {_s[k][1:]} ")
                                _x = isSavarna(env[_s[k][1:]], env[k])  # noqa: F405
                            else:
                                _x = isSavarna(_s[k], env[k])  # noqa: F405
                            logger.debug(f"Got {_x}")
                            x = x and _x
                        return x

                    for k in ["olp", "orp", "lp", "rp"]:
                        if k in s.keys():
                            # Set or remove one tag
                            def _tag(k, sk):
                                if sk[0:2] == "++":
                                    logger.debug(f"Setting {k} it {sk[2:]}")
                                    env[k].setIt(sk[2:])
                                elif sk[0:2] == "--":
                                    logger.debug(f"Removing {k} it {sk[1:]}")
                                    env[k].deleteIt(sk[2:])
                                elif sk[0] == "+":
                                    logger.debug(f"Setting {k} tag {sk[1:]}")
                                    env[k].setTag(sk[1:])
                                elif sk[0] == "-":
                                    logger.debug(f"Removing {k} tag {sk[1:]}")
                                    if env[k].hasTag(sk[1:]):
                                        env[k].deleteTag(sk[1:])
                                elif sk == "_lu":
                                    logger.debug("Removing all tags ")
                                    env[k].luTags()
                                elif sk[0] == "=":  # Replace
                                    logger.debug(f"Replacing {k} with {sk[1:]} {eval(sk[1:])}")
                                    env[k] = eval(sk[1:])  # Must be defined!

                            # Possibly set/remove multiple tags
                            if isinstance(s[k], list):
                                for sk in s[k]:
                                    _tag(k, sk)
                            else:
                                _tag(k, s[k])

                    if "domain" in s.keys():
                        st = s["domain"]
                        for k in st:
                            logger.debug(f"Updating domain {k} {st[k]}")
                            cond = True
                            if "condition" in st[k]:
                                logger.debug(f"Update condition check {st[k]['condition']}")
                                # List implies an or in condition
                                if isinstance(st[k]['condition'], list):
                                    cond = False
                                    for _s in st[k]['condition']:
                                        cond = cond or _c(env)
                                else:
                                    _s = st[k]['condition']
                                    cond = _c(env)
                                logger.debug(f"Check got {cond}")
                            if cond:
                                setattr(domains, k, st[k]["value"])
                return _update
            supdate = _exec_update(s["update"])
        if s["id"] in sutra_dict:
            logger.error(f"Duplicate Sutra {s['id']} - {sutra_dict[s['id']]} and {sname}")
            assert False
        sutra_dict[s["id"]] = LRSutra(sname, s["id"],
                                      cond=scond,
                                      xform=sxform,
                                      insert=sinsert,
                                      domain=sdom,
                                      update=supdate,
                                      optional=sopt,
                                      bahiranga=s["bahiranga"],
                                      overrides=soverrides)

    return sutra_dict



from sanskrit_parser.generator.paninian_object import PaninianObject
from sanskrit_parser.base.sanskrit_base import SLP1


class Dhatu(PaninianObject):
    """ Sanskrit Object Class: Derived From SanskritString

     Attributes:
    """
    def __init__(self, thing=None, its=[], other_tags=[], encoding=SLP1,
                 unicode_encoding='utf-8',
                 strict_io=True, replace_ending_visarga='s'):
        super().__init__(thing, encoding, unicode_encoding, strict_io, replace_ending_visarga)
        self.inPrakriya = True
        self.its = its
        self.setTag("DAtu")
        self.setTag("aNga")
        for t in other_tags:
            self.setTag(t)

    def hasIt(self, it):
        return it in self.its


iR = Dhatu("i", its=["R"], other_tags=["eti"])
eDa = Dhatu("eD", other_tags=["eDati"], its=["a"])
lUY = Dhatu("lU", its=["Y"])
kzI = Dhatu("kzI")
ji = Dhatu("ji")
wukrIY = Dhatu("krI", its=["Y", "wu"])
veY = Dhatu("veY", its=["Y"])
fcCa = Dhatu("fcC", its=["~a"])
Cad = Dhatu("Cad")
mud = Dhatu("mud")
vid = Dhatu("vid")
gfj = Dhatu("gfj")
BU = Dhatu("BU")
as_dhatu = Dhatu("as")
qulaBaz = Dhatu("laB", its=["~a", "z", "qu"])
guhU = Dhatu("guh", its=["~u"])
sTA = Dhatu("sTA", other_tags=["sTA"])

# FIXME: temporary for testing, samprasarana version of veY. Remove later
veY_smp = Dhatu("u", its=["Y"])



from sanskrit_parser.generator.paninian_object import PaninianObject
from sanskrit_parser.base.sanskrit_base import SLP1


class Pratipadika(PaninianObject):
    """ Sanskrit Object Class: Derived From SanskritString

     Attributes:
    """
    def __init__(self, thing=None, linga="pum",
                 other_tags=[], encoding=SLP1,
                 unicode_encoding='utf-8',
                 strict_io=True, replace_ending_visarga='s'):
        super().__init__(thing, encoding, unicode_encoding, strict_io, replace_ending_visarga)
        self.linga = linga
        self.inPrakriya = True
        self.setTag("prAtipadika")
        self.setTag(linga)
        # self.setTag("aNga")
        for t in other_tags:
            self.setTag(t)

    def anta(self):
        return self.canonical()[-1]


rAma = Pratipadika("rAma", "pum")
kavi = Pratipadika("kavi", "pum")
hari = Pratipadika("hari", "pum")
saKi = Pratipadika("saKi", "pum", other_tags=["saKi"])
pati = Pratipadika("pati", "pum", other_tags=["pati"])
pAda = Pratipadika("pAda", "pum",  other_tags=["pAdAdi"])
yUza = Pratipadika("yUza", "pum",  other_tags=["pAdAdi"])
sarva = Pratipadika("sarva", "pum", other_tags=["sarvAdi"])
krozwu = Pratipadika("krozwu", "pum")
SamBu = Pratipadika("SamBu", "pum")
go = Pratipadika("go", "pum")
rE = Pratipadika("rE", "pum")

# f
pitf = Pratipadika("pitf", "pum")
nf = Pratipadika("nf", "pum")
tvazwf = Pratipadika("tvazwf", "pum", other_tags=["naptrAdi"])
naptf = Pratipadika("naptf", "pum", other_tags=["naptrAdi"])
nezwf = Pratipadika("nezwf", "pum", other_tags=["naptrAdi"])
kzatf = Pratipadika("kzatf", "pum", other_tags=["naptrAdi"])
hotf = Pratipadika("hotf", "pum", other_tags=["naptrAdi"])
potf = Pratipadika("potf", "pum", other_tags=["naptrAdi"])
praSAstf = Pratipadika("praSAstf", "pum", other_tags=["naptrAdi"])


# क्विबन्ताः विजन्ताश्च प्रातिपदिकत्वं न. जहति, धातुत्वमपि न मुञ्चन्ति
viSvapA = Pratipadika("viSvapA", "pum", other_tags=["DAtu", "vic"])
senAnI = Pratipadika("senAnI", "pum", other_tags=["DAtu", "kvip"])
nI = Pratipadika("nI", "pum", other_tags=["DAtu", "kvip"])
KalapU = Pratipadika("KalapU", "pum", other_tags=["DAtu", "kvip"])
varzABU = Pratipadika("varzABU", "pum", other_tags=["DAtu", "BU", "kvip"])
svayamBU = Pratipadika("svayamBU", "pum", other_tags=["DAtu", "BU", "kvip"])
suDI = Pratipadika("suDI", "pum", other_tags=["DAtu", "kvip", "purvastrI"])

hAhA = Pratipadika("hAhA", "pum")

rAjan = Pratipadika("rAjan", "pum")
mahat = Pratipadika("mahat", "pum")

# FIXME - remove this, derive from qati

kati = Pratipadika("kati", "pum", other_tags=["qati", "nityabahuvacana"])

# saMKyA
tri = Pratipadika("tri", "pum", other_tags=["saMKyA", "nityabahuvacana"])
dvi = Pratipadika("dvi", "pum", other_tags=["saMKyA", "nityadvivacana",
                                            'tyadAdi'])
catur = Pratipadika("catur", "pum", other_tags=["saMKyA", "nityabahuvacana"])
# tri_s = Pratipadika("tri", "strI", other_tags=["saMKyA", "nityabahuvacana"])
# catur_s = Pratipadika("catur", "strI", other_tags=["saMKyA", "nityabahuvacana"])
dvi_s = Pratipadika("dvi", "strI", other_tags=["saMKyA", "nityadvivacana",
                                               'tyadAdi', "Ap"])

# Stri
ap = Pratipadika("ap", "strI", other_tags=["nityabahuvacana"])
mAtf = Pratipadika("mAtf", "strI", other_tags=["svasrAdi"])
svasf = Pratipadika("svasf", "strI", other_tags=["svasrAdi", "naptrAdi"])
tisf = Pratipadika("tisf", "strI", other_tags=["svasrAdi", "saMKyA",
                                               "nityabahuvacana"])
catasf = Pratipadika("catasf", "strI", other_tags=["svasrAdi""saMKyA",
                                                   "nityabahuvacana"])
nanAndf = Pratipadika("nanAndf", "strI", other_tags=["svasrAdi"])
duhitf = Pratipadika("duhitf", "strI", other_tags=["svasrAdi"])
yAtf = Pratipadika("yAtf", "strI", other_tags=["svasrAdi"])

BrU = Pratipadika("BrU", "strI", other_tags=["BrU"])
ramA = Pratipadika("ramA", "strI", other_tags=["Ap"])
nadI = Pratipadika("nadI", "strI", other_tags=["NI"])
niSA = Pratipadika("niSA", "strI", other_tags=["pAdAdi", "Ap"])
nAsikA = Pratipadika("nAsikA", "strI", other_tags=["pAdAdi", "Ap"])
mati = Pratipadika("mati", "strI")
lakzmI = Pratipadika("lakzmI", "strI")  # No NI
strI = Pratipadika("strI", "strI", other_tags=["NI", "strI_p"])
SrI = Pratipadika("SrI", "strI", other_tags=["DAtu", "kvip"])
# to test
Denu = Pratipadika("Denu", "strI")
suBrU = Pratipadika("suBrU", "strI", other_tags=["BrU"])

# Napum

jYAna = Pratipadika("jYAna", "napum")
anya = Pratipadika("anya", "napum", other_tags=["qatarAdi", "sarvanAma"])
anyatara = Pratipadika("anyatara", "napum", other_tags=["qatarAdi", "sarvanAma"])
itara = Pratipadika("itara", "napum", other_tags=["qatarAdi", "sarvanAma"])
qatara = Pratipadika("qatara", "napum", other_tags=["qatarAdi", "sarvanAma"])
qatama = Pratipadika("qatama", "napum", other_tags=["qatarAdi", "sarvanAma"])
vAri = Pratipadika("vAri", "napum")
mahat_n = Pratipadika("mahat", "napum")
payas = Pratipadika("payas", "napum")
SrIpA = Pratipadika("SrIpA", "napum", other_tags=["DAtu", "kvip"])
asTi = Pratipadika("asTi", "napum")
daDi = Pratipadika("daDi", "napum")
sakTi = Pratipadika("sakTi", "napum")
akzi = Pratipadika("akzi", "napum")
atirE = Pratipadika("atirE", "napum")
atinO = Pratipadika("atinO", "napum")
